"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Cobbler = exports.BaseCobbler = void 0;
const gdut_json_1 = require("gdut-json");
const gdut_io_1 = require("gdut-io");
function BaseCobbler() {
    const path = './cobbler';
    const tables = gdut_json_1.ReadJson(`${path}/tables.json`);
    return new Cobbler(`${path}/data`, tables);
}
exports.BaseCobbler = BaseCobbler;
class Cobbler {
    constructor(dataPath, tables) {
        this.dataPath = dataPath;
        this.tables = tables;
    }
    getTables() {
        return Object.entries(this.tables).map(([key, value]) => {
            return Object.assign({ tableName: key }, value);
        });
    }
    save(name, record) {
        const records = this.readDataFile(name);
        let saved = null;
        if (record.id && record.id > 0) {
            const found = this.locate(record.id, records);
            if (!found) {
                saved = this.insert(name, records, record);
            }
            else {
                if (found.logicRemove) {
                    throw new Error(`RECORD_REMOVED; Record "${found.id}" already Removed! 🤯`);
                }
                saved = this.update(name, records, found, record);
            }
        }
        else {
            saved = this.insert(name, records, record);
        }
        this.writeDataFile(name, records);
        return saved;
    }
    remove(name, id) {
        const records = this.readDataFile(name);
        const found = this.locate(id, records);
        if (found) {
            found.logicRemove = true;
            this.writeDataFile(name, records);
            return found;
        }
        else {
            throw new Error(`RECORD_NOT_FOUND; Record not found ${id}! 😵`);
        }
    }
    findById(name, id) {
        const records = this.readDataFile(name);
        const found = this.locate(id, records);
        if (found) {
            this.transformRecord(name, found);
            this.collectOwnedReferences(name, found);
            return found;
        }
        else {
            return null;
        }
    }
    findAll(name) {
        const records = this.readDataFile(name);
        return records
            .filter((el) => !el.logicRemove)
            .map((el) => {
            this.transformRecord(name, el);
            this.collectOwnedReferences(name, el);
            return el;
        });
    }
    findPage(name, page, size, sortName, sortDir) {
        const records = this.findAll(name);
        return records
            .sort(CompareValues(sortName, sortDir))
            .slice(page * size, page * size + size);
    }
    transformRecord(name, record, parent) {
        const tableConfig = this.tables[name];
        Object.keys(tableConfig).forEach((columnName) => {
            const columnConfig = tableConfig[columnName];
            const value = record[columnName];
            if (value !== null &&
                columnConfig.reference &&
                columnConfig.reference !== parent) {
                const refs = this.readDataFile(columnConfig.reference);
                record[columnName] = refs.find((el) => el.id === value);
            }
        });
        return record;
    }
    collectOwnedReferences(name, record) {
        Object.keys(this.tables)
            .filter((tableName) => tableName !== name)
            .forEach((tableName) => {
            const tableConfig = this.tables[tableName];
            Object.entries(tableConfig).filter(([colName, colConfig]) => {
                if (colConfig.reference === name && colConfig.owner) {
                    const refs = this.readDataFile(tableName);
                    record[tableName] = refs
                        .filter((ref) => !ref.logicRemove &&
                        ref[colName] === record.id)
                        .map((el) => {
                        this.transformRecord(tableName, el, name);
                        return el;
                    });
                }
            });
        });
    }
    update(name, records, found, record) {
        Object.keys(record).forEach((key) => {
            if (!['id', 'logicRemove'].includes(key)) {
                found[key] = record[key];
            }
        });
        this.validate(name, records, found, 'UPDATE');
        return found;
    }
    insert(name, records, record) {
        if (typeof record.id === 'undefined' || record.id < 1) {
            record.id = records.length + 1;
        }
        record.logicRemove = false;
        this.validate(name, records, record, 'INSERT');
        records.push(record);
        return record;
    }
    validate(name, records, record, action) {
        if (this.tables[name]) {
            const columns = this.tables[name];
            Object.keys(columns).forEach((columnName) => {
                const columnConfig = columns[columnName];
                const value = record[columnName];
                const valueType = typeof record[columnName];
                this.checkColumnType(valueType, columnConfig, record, columnName, value);
                this.checkReference(columnConfig, columnName, value);
                this.checkUnique(record, records, columnName, columnConfig, value);
            });
        }
    }
    checkUnique(record, records, columnName, columnConfig, value) {
        if (columnConfig.unique) {
            const found = records.find((el) => el.id !== record.id && el[columnName] === value);
            if (found) {
                throw new Error(`UNIQUE_CONFLICT; The "${columnName}" / [${value}] is not unique - id:${found.id}! 😬`);
            }
        }
    }
    checkColumnType(valueType, columnConfig, record, columnName, value) {
        if (valueType === 'undefined' || value === null) {
            if (typeof columnConfig.default !== 'undefined') {
                record[columnName] = columnConfig.default;
                valueType = typeof record[columnName];
            }
            else if (columnConfig.needed) {
                throw new Error(`COLUMN_NEEDED; Column needed: "${columnName}" ! 🤨`);
            }
            else {
                record[columnName] = null;
                value = null;
            }
        }
        if (valueType !== columnConfig.type && value !== null) {
            throw new Error(`COLUMN_TYPE_MISMATCH; Column "${columnName}" type "${columnConfig.type}" mismatch "${valueType}" ! 😲`);
        }
    }
    checkReference(columnConfig, columnName, value) {
        if (columnConfig.reference && value !== null) {
            if (this.tables[columnConfig.reference]) {
                const refRecs = this.readDataFile(columnConfig.reference);
                const refFound = refRecs.find((el) => el.id === value);
                if (!refFound) {
                    throw new Error(`Column "${columnName}" - reference record not found ${columnConfig.reference} / ${value}!`);
                }
                else if (refFound.logicRemove) {
                    throw new Error(`Column "${columnName}" - reference record deleted ${columnConfig.reference} / ${value}!`);
                }
            }
            else {
                throw new Error(`Column "${columnName}" - reference table not found: ${columnConfig.reference}!`);
            }
        }
    }
    locate(id, data) {
        const found = data.find((rec) => rec.id === id);
        if (found) {
            return found;
        }
        else {
            return null;
        }
    }
    readDataFile(name) {
        const filePath = `${this.dataPath}/${name}.json`;
        if (gdut_io_1.FileExists(filePath)) {
            return gdut_json_1.ReadJson(filePath);
        }
        else {
            return [];
        }
    }
    writeDataFile(name, data) {
        const filePath = `${this.dataPath}/${name}.json`;
        gdut_json_1.WriteJson(filePath, data, true);
    }
}
exports.Cobbler = Cobbler;
function CompareValues(key, order = 'ASC') {
    return function innerSort(a, b) {
        if (!a.hasOwnProperty(key) || !b.hasOwnProperty(key)) {
            // property doesn't exist on either object
            return 0;
        }
        const varA = typeof a[key] === 'string' ? a[key].toUpperCase() : a[key];
        const varB = typeof b[key] === 'string' ? b[key].toUpperCase() : b[key];
        let comparison = 0;
        if (varA > varB) {
            comparison = 1;
        }
        else if (varA < varB) {
            comparison = -1;
        }
        return order === 'DESC' ? comparison * -1 : comparison;
    };
}
